package cn.com.qmth.examcloud.core.oe.admin.service.impl;

import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.core.oe.admin.base.jpa.Searcher;
import cn.com.qmth.examcloud.core.oe.admin.base.jpa.SpecUtils;
import cn.com.qmth.examcloud.core.oe.admin.base.utils.CommonUtil;
import cn.com.qmth.examcloud.core.oe.admin.dao.*;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.*;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.AuditStatus;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.DisciplineType;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.FaceVerifyResult;
import cn.com.qmth.examcloud.core.oe.admin.service.*;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.examaudit.RedoAuditInfo;
import cn.com.qmth.examcloud.marking.api.MarkWorkCloudService;
import cn.com.qmth.examcloud.marking.api.request.AppendMarkWorkPaperReq;
import cn.com.qmth.examcloud.support.enums.FaceBiopsyScheme;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;

import java.util.Date;
import java.util.List;
import java.util.Optional;

/**
 * 考试记录审核相关接口
 *
 * @author: QMTH
 * @since: 2018/8/27
 */
@Service
public class ExamAuditServiceImpl implements ExamAuditService {

    @Autowired
    private ExamAuditRepo examAuditRepo;

    @Autowired
    private ExamScoreRepo examScoreRepo;

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordQuestionsService examRecordQuestionsService;

    @Autowired
    private ExamRecordForMarkingRepo examRecordForMarkingRepo;

    @Autowired
    private ExamScorePushQueueService examScorePushQueueService;

    @Autowired
    private ExamScoreObtainQueueService examScoreObtainQueueService;

    @Autowired
    private MarkWorkCloudService markWorkCloudService;

    @Autowired
    private ExamFaceLivenessVerifyRepo examFaceLivenessVerifyRepo;

    @Autowired
    private FaceBiopsyItemRepo faceBiopsyItemRepo;

    @Autowired
    private ExamFaceLiveVerifyRepo examFaceLiveVerifyRepo;

    @Autowired
    private ExamStudentFinalScoreService examStudentFinalScoreService;

    @Autowired
    private IllegallyTypeRepo illegallyTypeRepo;

    @Autowired
    private ExamStudentFinalScoreRepo examStudentFinalScoreRepo;

    /**
     * 活体检测失败自动审核
     */
    private static final String FACE_LIVENESS_VERIFY_AUTO_AUDIT = "FACE_LIVENESS_VERIFY_AUTO_AUDIT";

    /**
     * 无照片自动审核
     */
    private static final String NOPHOTO_VERIFY_AUTO_AUDIT = "NOPHOTO_VERIFY_AUTO_AUDIT";

    /**
     * 系统审核
     */
    private static final String AUDIT_USER_NAME = "SYSTEM";


    @Override
    public ExamAuditEntity getExamAuditByExamRecordDataId(Long examRecordDataId) {
        if (examRecordDataId == null) {
            return null;
        }
        Searcher searchers = new Searcher()
                .eq("examRecordDataId", examRecordDataId);
        Specification<ExamAuditEntity> spec = SpecUtils.buildSearchers(ExamAuditEntity.class, searchers.build());
        List<ExamAuditEntity> list = examAuditRepo.findAll(spec);
        if (list != null && list.size() > 0) {
            return list.get(0);
        }
        return null;
    }

    @Override
    @Transactional
    public void singleAudit(Long examRecordDataId, Boolean isPass, String disciplineDetail,
                            String disciplineType, User user, Long illegallyTypeId) {
        ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId, ExamRecordDataEntity.class);
        if (examRecordData == null) {
            return;
        }
        //已经审核过了，这里直接返回，如果需要重新审核，调用重审redoAudit方法
        if (examRecordData.getIsAudit() != null && examRecordData.getIsAudit()) {
            return;
        }

        //保存审核记录
        ExamAuditEntity examAudit = new ExamAuditEntity();
        examAudit.setExamRecordDataId(examRecordDataId);
        examAudit.setStatus(isPass ? AuditStatus.PASS : AuditStatus.UN_PASS);
        if (null != illegallyTypeId) {
            IllegallyTypeEntity illegallyType =
                    GlobalHelper.getEntity(illegallyTypeRepo, illegallyTypeId, IllegallyTypeEntity.class);
            if (null == illegallyType) {
                throw new StatusException("100001", "违纪类型数据有误");
            }
            disciplineType = illegallyType.getCode();
        }
        examAudit.setDisciplineType(disciplineType);
        examAudit.setDisciplineDetail(disciplineDetail);
        if (user != null) {
            examAudit.setUserId(user.getUserId().toString());
            examAudit.setAuditUserName(user.getDisplayName());
        }
        examAudit.setIllegallyTypeId(illegallyTypeId);

        examAuditRepo.save(examAudit);

        //保存考试记录
        examRecordData.setIsAudit(true);
        examRecordData.setIsIllegality(isPass ? false : true);
        examRecordData.setUpdateTime(new Date());
        examRecordDataRepo.save(examRecordData);

        doAuditPostProcesser(isPass, examRecordData);
    }

    @Override
    public void redoAudit(RedoAuditInfo redoAuditInfo, User user) {
        for (Long examRecordDataId : redoAuditInfo.getExamRecordDataIds()) {
            ExamAuditEntity examAudit = examAuditRepo.findByExamRecordDataId(examRecordDataId);
            if (examAudit == null) {
                examAudit = new ExamAuditEntity();
                examAudit.setCreationTime(new Date());
                examAudit.setExamRecordDataId(examRecordDataId);
            }
            examAudit.setAuditUserName(user.getDisplayName());
            examAudit.setUserId(user.getUserId().toString());
            if (StringUtils.isBlank(redoAuditInfo.getDisciplineDetail())) {
                examAudit.setDisciplineDetail("重审");
            } else {
                examAudit.setDisciplineDetail(redoAuditInfo.getDisciplineDetail());
            }

            String disciplineType = redoAuditInfo.getDisciplineType();
            if (null != redoAuditInfo.getIllegallyTypeId()) {
                IllegallyTypeEntity illegallyType =
                        GlobalHelper.getEntity(illegallyTypeRepo, redoAuditInfo.getIllegallyTypeId(), IllegallyTypeEntity.class);
                if (null == illegallyType) {
                    throw new StatusException("100002", "违纪类型数据有误");
                }
                disciplineType = illegallyType.getCode();
            }

            examAudit.setDisciplineType(disciplineType);
            examAudit.setStatus(redoAuditInfo.getIsPass() ? AuditStatus.PASS : AuditStatus.UN_PASS);
            examAudit.setUpdateTime(new Date());

            examAudit.setIllegallyTypeId(redoAuditInfo.getIllegallyTypeId());

            examAuditRepo.save(examAudit);

            ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId, ExamRecordDataEntity.class);
            examRecordData.setIsWarn(true);
            examRecordData.setIsAudit(true);
            examRecordData.setIsIllegality(!redoAuditInfo.getIsPass());//是否违纪
            examRecordData.setUpdateTime(new Date());
            examRecordDataRepo.save(examRecordData);

            doAuditPostProcesser(redoAuditInfo.getIsPass(), examRecordData, redoAuditInfo.getFromMarking());
        }
    }

    /**
     * 活体检测失败，系统审核
     *
     * @param examRecordDataId
     * @param rootOrgId
     */
    @Override
    public void saveExamAuditByFaceVerifyFailed(Long examRecordDataId, Long rootOrgId) {
        boolean exist = examAuditRepo.existsByExamRecordDataId(examRecordDataId);
        if (exist) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(examRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.ACTION_FAILURE.name());
        examAuditEntity.setUserId(FACE_LIVENESS_VERIFY_AUTO_AUDIT);//活体检测自动审核
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setDisciplineDetail(getFaceVerifyDisciplineDetail(examRecordDataId, rootOrgId));//计算违纪描述
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);

        // 修改考试记录为违纪、异常、已审
        this.disciplineExamRecordData(examRecordDataId);
    }

    /**
     * 抓拍照片为0，系统审核
     */
    @Override
    public void saveExamAuditByNoPhoto(Long examRecordDataId) {
        boolean exist = examAuditRepo.existsByExamRecordDataId(examRecordDataId);
        if (exist) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(examRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.OTHER.name());
        examAuditEntity.setDisciplineDetail("抓拍照片数量为0");
        examAuditEntity.setUserId(NOPHOTO_VERIFY_AUTO_AUDIT);//抓拍照片为0自动审核
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);

        // 修改考试记录为违纪、异常、已审
        this.disciplineExamRecordData(examRecordDataId);
    }

    @Override
    public void saveExamAuditForIllegalClient(Long realExamRecordDataId, String reason) {
        boolean exist = examAuditRepo.existsByExamRecordDataId(realExamRecordDataId);
        if (exist) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(realExamRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.ILLEGAL_CLIENT.name());
        if (StringUtils.isNotEmpty(reason)) {
            // 限制字符串最大长度
            reason = StringUtils.substring(reason, 0, 1000);
            examAuditEntity.setDisciplineDetail(reason);
        }
        examAuditEntity.setUserId(DisciplineType.ILLEGAL_CLIENT.name());
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);

        // 修改考试记录为违纪、异常、已审
        this.disciplineExamRecordData(realExamRecordDataId);
    }

    @Override
    public void saveHeaderWarnAudit(Long realExamRecordDataId) {
        boolean exist = examAuditRepo.existsByExamRecordDataId(realExamRecordDataId);
        if (exist) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(realExamRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.ILLEGAL_DATA.name());
        examAuditEntity.setDisciplineDetail(null);
        examAuditEntity.setUserId(DisciplineType.ILLEGAL_DATA.name());
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);

        // 修改考试记录为违纪、异常、已审
        this.disciplineExamRecordData(realExamRecordDataId);
    }

    private void disciplineExamRecordData(Long realExamRecordDataId) {
        Optional<ExamRecordDataEntity> optional = examRecordDataRepo.findById(realExamRecordDataId);
        if (optional.isPresent()) {
            // 修改考试记录为违纪、异常、已审
            ExamRecordDataEntity entity = optional.get();
            entity.setIsIllegality(true);
            entity.setIsWarn(true);
            entity.setIsAudit(true);
            examRecordDataRepo.save(entity);
        }
    }

    //获取活检违纪详情
    private String getFaceVerifyDisciplineDetail(Long examRecordDataId, Long rootOrgId) {
        FaceBiopsyScheme faceBiopsyScheme = FaceBiopsyHelper.getFaceBiopsyScheme(rootOrgId);
        if (FaceBiopsyScheme.FACE_CLIENT == faceBiopsyScheme) {
            // C端活体检测方案
            List<ExamFaceLiveVerifyEntity> entities = examFaceLiveVerifyRepo.findByExamRecordDataId(examRecordDataId);
            if (CollectionUtils.isEmpty(entities)) {
                return "未进行人脸活体检测";
            }

            StringBuilder result = new StringBuilder();
            for (ExamFaceLiveVerifyEntity entity : entities) {
                result.append(CommonUtil.getDateStrWithSecond(entity.getCreationTime()))
                        .append(":")
                        .append(entity.getStatus().getDescription())
                        .append("&&");
            }

            return result.toString();
        } else if (faceBiopsyScheme == FaceBiopsyScheme.FACE_MOTION) {
            // Electron Client 自研活体检测方案
            List<FaceBiopsyItemEntity> faceBiopsyItems = faceBiopsyItemRepo.findByExamRecordDataIdOrderByIdAsc(examRecordDataId);

            //从未调过人脸检测接口，或者只调过接口但是未保存过结果(即未完成活检)，则提示未进行人脸活体检测
            if (faceBiopsyItems == null || (faceBiopsyItems.size() == 1 && !faceBiopsyItems.get(0).getCompleted())) {
                return "未进行人脸活体检测";
            }

            StringBuffer sbBuffer = new StringBuffer("");
            for (FaceBiopsyItemEntity item : faceBiopsyItems) {
                String startTime = CommonUtil.getDateStrWithSecond(item.getCreationTime());
                String resultMsg = "";
                if (null != item.getCompleted() && !item.getCompleted()) {
                    resultMsg = "活检未完成";
                } else {
                    if (null != item.getResult() && item.getResult()) {
                        resultMsg = "活检成功";
                    } else {
                        resultMsg = item.getErrorMsg() == null ? "" : item.getErrorMsg();
                    }
                }
                sbBuffer.append(startTime + ":" + resultMsg);
                sbBuffer.append("&&");
            }
            return sbBuffer.toString();
        } else {
            // FaceID活体检测方案
            List<ExamFaceLivenessVerifyEntity> faceVerifies =
                    examFaceLivenessVerifyRepo.findByExamRecordDataIdOrderById(examRecordDataId);
            if (faceVerifies == null || faceVerifies.size() == 0) {
                return "未进行人脸动作检测";
            } else {
                StringBuffer sbBuffer = new StringBuffer("");
                for (ExamFaceLivenessVerifyEntity faceVerify : faceVerifies) {
                    String startTime = CommonUtil.getDateStrWithSecond(faceVerify.getStartTime());
                    sbBuffer.append(startTime + ":" + FaceVerifyResult.getDescByName(faceVerify.getVerifyResult()));
                    sbBuffer.append("&&");
                }
                String disciplineDetail = sbBuffer.toString();
                if (disciplineDetail.length() > 500) {
                    disciplineDetail = disciplineDetail.substring(0, 500);
                }
                return disciplineDetail;
            }
        }
    }

    /**
     * 重审后置处理
     *
     * @param isPass
     * @param examRecordData
     */
    private void doAuditPostProcesser(Boolean isPass, ExamRecordDataEntity examRecordData, Boolean fromMarking) {
        if (isPass) {
            auditPassPostProcesser(examRecordData);
        } else {
            auditUnPassPostProcesser(examRecordData, fromMarking);
        }

        //重新计算考生的最终分数
        examStudentFinalScoreService.calcAndSaveFinalScore(examRecordData.getExamStudentId());
    }

    private void doAuditPostProcesser(Boolean isPass, ExamRecordDataEntity examRecordData) {
        doAuditPostProcesser(isPass, examRecordData, null);
    }

    /**
     * 审核通过后置处理
     */
    private void auditPassPostProcesser(ExamRecordDataEntity examRecordData) {
        ExamScoreEntity examScore = examScoreRepo.findByExamRecordDataId(examRecordData.getId());
        //审核通过且为非全客观题卷，保存阅卷需要数据
        saveExamRecordForMarkingWithAudit(examRecordData, examScore);
        //保存推送分数队列
        examScorePushQueueService.saveScoreDataInfoToQueue(examRecordData.getId(), examScore.getId());
        //保存获取分数队列
        examScoreObtainQueueService.saveExamScoreObtainQueue(examRecordData.getId());
        //发送通知
        examScoreObtainQueueService.sendObtainScoreNotify(examRecordData.getRootOrgId());
    }

    /**
     * 审核通过且为非全客观题卷，保存阅卷需要数据
     *
     * @param examRecordData
     */
    private void saveExamRecordForMarkingWithAudit(ExamRecordDataEntity examRecordData, ExamScoreEntity examScore) {
        if (examRecordData.getIsAllObjectivePaper() != null && !examRecordData.getIsAllObjectivePaper()) {
            ExamRecordForMarkingEntity examRecordForMarkingExists =
                    examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
            if (examRecordForMarkingExists != null) {
                return;
            }
            ExamRecordForMarkingEntity examRecordForMarking = new ExamRecordForMarkingEntity();
            examRecordForMarking.setExamId(examRecordData.getExamId());
            examRecordForMarking.setExamRecordDataId(examRecordData.getId());
            examRecordForMarking.setExamStudentId(examRecordData.getExamStudentId());
            examRecordForMarking.setBasePaperId(examRecordData.getBasePaperId());
            examRecordForMarking.setPaperType(examRecordData.getPaperType());
            examRecordForMarking.setCourseId(examRecordData.getCourseId());
            examRecordForMarking.setObjectiveScore(examScore.getObjectiveScore());
            int subjectiveAnswerLength = examRecordQuestionsService.calculationSubjectiveAnswerLength(examRecordData.getId());
            examRecordForMarking.setSubjectiveAnswerLength(subjectiveAnswerLength);
            examRecordForMarkingRepo.save(examRecordForMarking);

            //调用阅卷接口，通知有新数据
            AppendMarkWorkPaperReq appendMarkWorkPaperReq = new AppendMarkWorkPaperReq();
            appendMarkWorkPaperReq.setRootOrgId(examRecordData.getRootOrgId());
            appendMarkWorkPaperReq.setExamId(examRecordData.getExamId());
            markWorkCloudService.appendMarkWorkPaper(appendMarkWorkPaperReq);
        }
    }

    /**
     * 审核不通过后置处理
     *
     * @param examRecordData
     */
    private void auditUnPassPostProcesser(ExamRecordDataEntity examRecordData, Boolean fromMarking) {
        if (fromMarking == null || !fromMarking) {
            ExamRecordForMarkingEntity examRecordForMarking = examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
            if (examRecordForMarking != null) {
                //删除阅卷需要数据，防止违纪数据进入阅卷
                examRecordForMarkingRepo.delete(examRecordForMarking);
            }
        }

        //删除考生的最终分数，后续会重新计算
        ExamStudentFinalScoreEntity finalExamScore =
                examStudentFinalScoreRepo.findByExamStudentId(examRecordData.getExamStudentId());
        if (finalExamScore != null) {
            examStudentFinalScoreRepo.delete(finalExamScore);
        }

        //保存获取分数队列
        examScoreObtainQueueService.saveExamScoreObtainQueue(examRecordData.getId());
        //发送通知
        examScoreObtainQueueService.sendObtainScoreNotify(examRecordData.getRootOrgId());
    }

    @Override
    public void saveExceedMaxSwitchScreenCount(Long realExamRecordDataId) {
        boolean exist = examAuditRepo.existsByExamRecordDataId(realExamRecordDataId);
        if (exist) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(realExamRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.EXCEED_MAX_SWITCH_SCREEN_COUNT.name());
        examAuditEntity.setDisciplineDetail(null);
        examAuditEntity.setUserId(DisciplineType.EXCEED_MAX_SWITCH_SCREEN_COUNT.name());
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);

        // 修改考试记录为违纪、异常、已审
        this.disciplineExamRecordData(realExamRecordDataId);
    }

    @Override
    public boolean updateExamAuditByAllPass(Long examRecordDataId) {
        ExamAuditEntity examAuditEntity = examAuditRepo.findByExamRecordDataId(examRecordDataId);
        if (examAuditEntity == null) {
            return false;
        }
        if (AuditStatus.PASS.equals(examAuditEntity.getStatus())) {
            return true;
        }
        examAuditEntity.setDisciplineType(null);
        examAuditEntity.setDisciplineDetail("开启审核全通过");
        examAuditEntity.setStatus(AuditStatus.PASS);
        examAuditEntity.setIllegallyTypeId(null);
        examAuditRepo.save(examAuditEntity);
        return true;
    }

}
